/*
 * Copyright (c) 2024 Google LLC.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/drivers/display.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/kernel.h>

namespace
{

// TODO: michaelh: pull from DeviceTree.
constexpr int kWidth = 128;
constexpr int kHeight = 64;

template <int Width, int Height> struct Frame {
	static constexpr int kWidth = Width;
	static constexpr int kHeight = Height;

	void set(int x, int y);
	void clear(int x, int y);
	bool get(int x, int y);

	uint8_t bits[kWidth * kHeight / 8] = {};
};

template <int Width, int Height> void Frame<Width, Height>::set(int x, int y)
{
	int offset = x + (y >> 3) * kWidth;
	int bit = y & 7;
	bits[offset] |= 1 << bit;
}

template <int Width, int Height>
void Frame<Width, Height>::clear(int x, int y)
{
	int offset = x + (y >> 3) * kWidth;
	int bit = y & 7;
	bits[offset] &= ~(1 << bit);
}

template <int Width, int Height> bool Frame<Width, Height>::get(int x, int y)
{
	int offset = x + (y >> 3) * kWidth;
	int bit = y & 7;
	return (bits[offset] >> bit) & 1;
}

uint32_t minrand(uint32_t &state)
{
	constexpr uint32_t a = 48271;

	uint32_t low = (state & 0x7fff) * a;
	uint32_t high = (state >> 15) * a;
	uint32_t x = low + ((high & 0xffff) << 15) + (high >> 16);

	state = (x & 0x7fffffff) + (x >> 31);
	return state;
}

template <typename F> void set_random(F &frame, uint32_t &state)
{
	frame.set(minrand(state) % (frame.kWidth - 1),
		  1 + minrand(state) % (frame.kHeight - 2));
}

template <typename F> void fill(F &frame, uint32_t &state)
{
	for (int i = 0; i < frame.kWidth * frame.kHeight / 10; ++i) {
		set_random(frame, state);
	}
}

template <typename F> int update(F &frame)
{
	Frame<1, kHeight> left;
	int spawned = 0;

	for (int x = 0; x < frame.kWidth - 1; ++x) {
		Frame<1, kHeight> current;
		for (int y = 0; y < current.kHeight / 8; ++y) {
			current.bits[y] = frame.bits[y * frame.kWidth + x];
		}

		for (int y = 1; y < frame.kHeight - 1; ++y) {
			int neighbours =
				// Top
				left.get(0, y - 1) + current.get(0, y - 1) +
				frame.get(x + 1, y - 1)
				// Mid
				+ left.get(0, y) +
				frame.get(x + 1, y)
				// Bottom
				+ left.get(0, y + 1) + current.get(0, y + 1) +
				frame.get(x + 1, y + 1);
			if (current.get(0, y)) {
				if (neighbours < 2 || neighbours > 3) {
					frame.clear(x, y);
				}
			} else {
				if (neighbours == 3) {
					frame.set(x, y);
					++spawned;
				}
			}
		}
		left = current;
	}
	return spawned;
}

int conway()
{
	Frame<kWidth, kHeight> fb;
	const device *dev = DEVICE_DT_GET(DT_CHOSEN(zephyr_display));
	struct display_buffer_descriptor fb_desc = {
		.buf_size = sizeof(fb.bits),
		.width = fb.kWidth,
		.height = fb.kHeight,
		.pitch = fb.kWidth,
	};

	uint32_t state = 1337;
	fill(fb, state);

	for (;;) {
		display_write(dev, 0, 0, &fb_desc, fb.bits);
		if (update(fb) < 20) {
			set_random(fb, state);
		}
	}
	return 0;
}

} // namespace

int main(void)
{
	conway();
	return 0;
}
