/*
 * Copyright (c) 2024 Google LLC.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/drivers/display.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/kernel.h>

namespace
{

constexpr int kWidth = 160;
constexpr int kHeight = 80;

constexpr uint8_t kKite[] = {
	0x00, 0x00, 0x00, 0x00, 0x7f, 0xe0, 0x00, 0xff, 0xf0, 0x00, 0xf0,
	0xf0, 0x00, 0xd9, 0xf0, 0x01, 0x8f, 0xb0, 0x01, 0x87, 0x30, 0x01,
	0x8f, 0x30, 0x01, 0x1f, 0xb0, 0x03, 0x38, 0xf0, 0x03, 0x70, 0x70,
	0x03, 0xe0, 0x70, 0x07, 0xc3, 0xe0, 0x07, 0xbf, 0x80, 0x07, 0xfc,
	0x00, 0x0f, 0xc0, 0x00, 0x1c, 0x01, 0xf8, 0x38, 0x03, 0xfc, 0x30,
	0x07, 0x0e, 0x60, 0x0e, 0x06, 0x60, 0x1c, 0x06, 0x7f, 0xf8, 0x06,
	0x3f, 0xe0, 0x06, 0x00, 0x00, 0x00,
};

constexpr int kImageWidth = 24;

constexpr uint16_t rgb565(uint8_t r, uint8_t g, uint8_t b)
{
	uint16_t colour = ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3);
	return (colour >> 8) | (colour << 8);
}

constexpr uint16_t kSkyBlue = rgb565(0x87, 0xce, 0xeb);
constexpr uint16_t kWhite = rgb565(0xFF, 0xFF, 0xFF);

uint32_t minrand(uint32_t &state)
{
	constexpr uint32_t a = 48271;

	uint32_t low = (state & 0x7fff) * a;
	uint32_t high = (state >> 15) * a;
	uint32_t x = low + ((high & 0xffff) << 15) + (high >> 16);

	state = (x & 0x7fffffff) + (x >> 31);
	return state;
}

int next_velocity(int current, int scale, uint32_t &state)
{
	int absolute = 1 + (minrand(state) % (scale - 1));
	return (current < 0) ? absolute : -absolute;
}

void render(const device *dev, int x, int y, uint16_t width, uint16_t height,
	    const uint8_t *bitmap, uint16_t white, uint16_t black,
	    bool mirror)
{
	uint16_t buffer[kImageWidth];
	struct display_buffer_descriptor descriptor = {
		.buf_size = sizeof(buffer),
		.width = width,
		.height = 1,
		.pitch = width,
	};
	for (int row = 0; row < height; ++row) {
		uint16_t *into = buffer;
		for (int col = 0; col < width; col += 8) {
			if (mirror) {
				uint8_t mask = bitmap[(width - col - 1) / 8];
				for (uint8_t bit = 1; bit != 0; bit <<= 1) {
					*into++ = ((mask & bit) != 0) ? white
								      : black;
				}
			} else {
				uint8_t mask = bitmap[col / 8];
				for (uint8_t bit = 0x80; bit != 0;
				     bit >>= 1) {
					*into++ = ((mask & bit) != 0) ? white
								      : black;
				}
			}
		}
		bitmap += width / 8;
		display_write(dev, x, y + row, &descriptor, buffer);
	}
}

int logogame()
{
	const device *dev = DEVICE_DT_GET(DT_CHOSEN(zephyr_display));
	uint32_t state = 1337;

	for (int x = 0; x < kWidth; x += kImageWidth) {
		for (int y = 0; y < kHeight; y += kImageWidth) {
			render(dev, x, y, kImageWidth, kImageWidth, kKite,
			       kSkyBlue, kSkyBlue, false);
		}
	}

	int x = 0;
	int y = 0;
	constexpr int kScale = 2;
	int dx = kScale;
	int dy = kScale - 1;

	for (int run = 0;; ++run) {
		render(dev, x / kScale, y / kScale, kImageWidth, kImageWidth,
		       kKite, kWhite, kSkyBlue, dx < 0);
		x += dx;
		if (x >= ((kWidth - kImageWidth) * kScale) || x < 1) {
			x -= dx;
			dx = next_velocity(dx, kScale, state);
		}
		y += dy;
		if (y >= ((kHeight - kImageWidth) * kScale) || y < 1) {
			y -= dy;
			dy = next_velocity(dy, kScale, state);
		}
		//		k_sleep(K_MSEC(10));
	}
	return 0;
}

} // namespace

int main(void)
{
	logogame();
	return 0;
}
